
import os 
from .dataset import create_dataset
from .models import SimpleClassificationModel, BERTClassificationModel
from .utils import get_tokenizer
from .configs import create_configs
import configparser

class TrainStrategy:
  def __init__(self, config_path):

    config = configparser.ConfigParser()
    config.read(config_path)

    data_config = configparser.ConfigParser()
    rel_path = os.path.dirname(__file__)
    data_ini_path = os.path.join(rel_path, "datasets.ini")
    data_config.read(data_ini_path)

    self.train_config, self.model_config = create_configs(config, data_config)
    self.datasets = create_dataset(config, data_config)

    if 'bert' in self.model_config['model_name']:
      self.model = BERTClassificationModel(self.model_config)
    else:
      self.model = SimpleClassificationModel(self.model_config)

  def start(self):
    self.model.train(self.datasets, **self.train_config)



import tnkeeh as tn 
from datasets import load_dataset
import os 
from .utils import get_preprocessing_args
from transformers import AutoTokenizer
import torch
from .utils import get_tokenizer


def split_dataset(dataset):
    if not('test' in dataset):
        train_valid_dataset = dataset['train'].train_test_split(test_size=0.1)
        train_valid_dataset['valid'] = train_valid_dataset.pop('test')

        train_test_dataset = train_valid_dataset['train'].train_test_split(test_size=0.1)
        
        dataset['train'] = train_test_dataset['train']
        dataset['test'] = train_test_dataset['test']
        dataset['valid'] = train_valid_dataset['valid']

    elif not('valid' in dataset):
        train_valid_dataset = dataset['train'].train_test_split(test_size=0.1)
        train_valid_dataset['valid'] = train_valid_dataset.pop('test')
        dataset['train'] = train_valid_dataset['train']
        dataset['valid'] = train_valid_dataset['valid']
        dataset['test'] = dataset['test']
    return dataset 


def clean_dataset(dataset, config, data_config):
    dataset_name = config['dataset']['dataset_name']
    text = data_config[dataset_name]['text']

    args = get_preprocessing_args(config)
    cleaner = tn.Tnkeeh(**args)
    dataset = cleaner.clean_hf_dataset(dataset, text)
    return dataset 

def write_data_for_train(dataset, text):
    data = []
    for sample in dataset:
        data.append(sample[text])    
    open(f'data.txt', 'w').write(('\n').join(data))

def create_dataset(config, data_config):

    tokenizer_name = config['tokenization']['tokenizer_name']
    max_tokens = int(config['tokenization']['max_tokens'])
    vocab_size  = int(config['tokenization']['vocab_size'])

    batch_size = int(config['train']['batch_size'])
    dataset_name = config['dataset']['dataset_name']
    model_name = config['model']['model_name']

    # clean and load data
    dataset = load_dataset(dataset_name)
    dataset = clean_dataset(dataset, config, data_config)

    # tokenize data
    if 'bert' in model_name:
        tokenizer = AutoTokenizer.from_pretrained(model_name, do_lower_case=False)
        dataset = dataset.map(lambda examples:tokenizer(examples['text'], truncation=True, padding='max_length'), batched=True)
        columns=['input_ids', 'token_type_ids', 'attention_mask', 'labels']
    else:
        write_data_for_train(dataset['train'], data_config[dataset_name]['text'])
        tokenizer = get_tokenizer(tokenizer_name)
        tokenizer = tokenizer(vocab_size = vocab_size)
        tokenizer.train('data.txt')
        dataset = dataset.map(lambda examples:{'input_ids': tokenizer.encode_sentences(examples['text'], out_length= max_tokens)}, batched=True)
        columns=['input_ids', 'labels']        
    
    # split datasets 
    dataset = dataset.map(lambda examples:{'labels': examples['label']}, batched=True)
    splits = split_dataset(dataset)
    
    #create loaders 
    for split in splits:
        dataset[split].set_format(type='torch', columns=columns)
        dataset[split] = torch.utils.data.DataLoader(dataset[split], batch_size=batch_size)
    return [dataset['train'], dataset['valid'], dataset['test']]



from transformers import AutoModelForSequenceClassification, AutoConfig
import os
import time
from tqdm import tqdm
import torch
from torch.optim import AdamW
from sklearn.metrics import accuracy_score
import torch.nn as nn

class BiRNN(nn.Module):
    def __init__(self, vocab_size, num_labels):
        
        super().__init__()
        
        self.embedding = nn.Embedding(vocab_size, 128)
        self.bigru = nn.GRU(128, 256, bidirectional=True)
        self.fc = nn.Linear(512, num_labels)
        self.num_labels = num_labels
        
    def forward(self, 
                input_ids,
                labels):

        embedded = self.embedding(input_ids)        
        out = self.bigru(embedded)
        logits = self.fc(out[0][:,0,:])
        loss = self.compute_loss(logits, labels)
        return {'loss':loss,
                'logits':logits} 
    
    def compute_loss(self, logits, labels):
        loss_fct = nn.CrossEntropyLoss()
        loss = loss_fct(logits.view(-1, self.num_labels), labels)
        return loss

class BaseModel:
    def __init__(self, config):
        self.model = nn.Module()
        self.num_labels = config['num_labels']
        self.model_name = config['model_name']
        self.vocab_size = config['vocab_size']

        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    def train(self, datasets, epochs = 30, save_dir = '.'):
        train_dataset, valid_dataset, test_dataset = datasets 
        filepath = os.path.join(save_dir, 'model.pth')
        best_accuracy = 0 
        for epoch in range(epochs):
            accuracy = 0 
            loss = 0 
            self.model.train().to(self.device)
            for _, batch in enumerate(train_dataset):
                batch = {k: v.to(self.device) for k, v in batch.items()}
                outputs = self.model(**batch)
                loss = outputs['loss']
                loss.backward()
                self.optimizer.step()
                self.optimizer.zero_grad()

                labels = batch['labels'].cpu() 
                preds = outputs['logits'].argmax(-1).cpu() 
                accuracy += accuracy_score(labels, preds) /len(train_dataset)
                loss += loss / len(train_dataset)
            
            print(f"Epoch {epoch} Train Loss {loss:.4f} Train Accuracy {accuracy:.4f}")
            
            self.model.eval().to(self.device)
            results = self.evaluate_dataset(valid_dataset)
            print(f"Epoch {epoch} Valid Loss {results['loss']:.4f} Valid Accuracy {results['accuracy']:.4f}")

            val_accuracy = results['accuracy']
            if val_accuracy > best_accuracy:
                best_accuracy = val_accuracy
                torch.save(self.model.state_dict(), filepath)

            #Later to restore:
        
        self.model.load_state_dict(torch.load(filepath))
        self.model.eval()
        results = self.evaluate_dataset(test_dataset)
        print(f"Test Loss {results['loss']:.4f} Test Accuracy {results['accuracy']:.4f}")
    
    def evaluate_dataset(self, dataset):
        accuracy = 0
        loss = 0 
        for _, batch in enumerate(dataset):
            batch = {k: v.to(self.device) for k, v in batch.items()}
            outputs = self.model(**batch)
            loss = outputs['loss']
            labels = batch['labels'].cpu() 
            preds = outputs['logits'].argmax(-1).cpu() 
            accuracy += accuracy_score(labels, preds) /len(dataset)
            loss += loss / len(dataset)
        return {'loss':loss, 'accuracy':accuracy}

class SimpleClassificationModel(BaseModel):
    def __init__(self, config):
        BaseModel.__init__(self, config)
        self.model = BiRNN(self.vocab_size, self.num_labels)
        self.model.to(self.device)    
        self.optimizer = AdamW(self.model.parameters(), lr=5e-5)

class BERTClassificationModel(BaseModel):
    def __init__(self, config):
        BaseModel.__init__(self, config)
        config = AutoConfig.from_pretrained(self.model_name,num_labels=self.num_labels)
        self.model = AutoModelForSequenceClassification.from_pretrained(self.model_name, config = config)
        self.optimizer = AdamW(self.model.parameters(), lr=5e-5)


def create_configs(config, data_config):
    dataset_name = config['dataset']['dataset_name']
    train_config = {'epochs':int(config['train']['epochs']),
                    'save_dir':config['train']['save_dir']}

    model_config = {'model_name':config['model']['model_name'],
                    'vocab_size':int(config['tokenization']['vocab_size']),
                    'num_labels':int(data_config[dataset_name]['num_labels'])}
    return train_config, model_config


import os
from setuptools import setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(name='nmatheg',
      version='0.0.2',
      url='',
      discription="Arabic Training Strategy For NLP Models",
      long_description=readme,
      long_description_content_type='text/markdown',
      author='Zaid Alyafeai, Maged Saeed',
      author_email='arabicmachinelearning@gmail.com',
      license='MIT',
      packages=['nmatheg'],
      install_requires=required,
      python_requires=">=3.6",
      include_package_data=True,
      zip_safe=False,
      )



import tkseem as tk

def get_tokenizer(tok_name):
    tokenizers = {'SentencePieceTokenizer':tk.SentencePieceTokenizer, 'WordTokenizer':tk.WordTokenizer, 'CharacterTokenizer':tk.CharacterTokenizer,
         'MorphologicalTokenizer':tk.MorphologicalTokenizer, 'RandomTokenizer':tk.RandomTokenizer, 'DisjointLetterTokenizer':tk.DisjointLetterTokenizer}
    return tokenizers[tok_name] 

def get_preprocessing_args(config):
    args = {}
    map_bool = {'True':True, 'False':False, '[]': []}
    for key in config['preprocessing']:
        val = config['preprocessing'][key]
        args[key] = map_bool[val]
    return args